from jellyfish import jaro_winkler, levenshtein_distance
from unicodedata import normalize
from string import punctuation
from functools import lru_cache


def remover_acentos(frase: str):
    """
    Remove acentos da frase parametro
    :param frase: parametros a remover acentos
    :return:
    """
    return normalize("NFKD", frase).encode("ASCII", "ignore").decode("ASCII")


@lru_cache(maxsize=4096)
def carregar_punctuation_dict():
    """
    Usada para carregar pontuacoes
    :return:
    """
    return {key: " " for key in punctuation}


def normalizar(frase: str, sort: bool = True):
    """
    Remove acentos/acentuacao, ordena as palavras e limpa espaços extras
    :param frase: frase a ser normalizada
    :param sort: true se as palavras devem ser ordenadas
    :return:
    """
    if frase:
        # remover acentos
        frase = remover_acentos(frase.lower())
        # remover acentuacao
        frase = frase.translate(str.maketrans(carregar_punctuation_dict()))
        # ordenar palavras
        if sort:
            frase = "".join(sorted(frase.split()))
        # limpar espaços extras
        frase = frase.strip()
    return frase


def comparador(org: str, dest: str, return_value: bool = False, porcentagem: int = 0.95):
    """
    Compara duas strings (org e dest) usando os algoritmos de Jaro–Winkler e distância Levenshtein
    :param org: string ser comparada
    :param dest: outra string ser comparada
    :param return_value: indica se o valor calculado deve ser retornado para uso posterior
    :param porcentagem: porcentagem usada na comparacao com o valor calculado (caso o valor nao precise ser retornado)
    :return:
    """
    jws: float = 0.0
    lvd: float = 0.0
    calc: float = 0.0

    if org and dest:
        jws = jaro_winkler(org, dest)
        lvd = abs(1 - (levenshtein_distance(org, dest) / 100))
        calc = round(((jws * 0.9) + (lvd * 0.1)), 2)
    if return_value:
        return calc
    return calc > porcentagem


def comparar_e_obter_mais_semelhante(org: str, dests: list, porcentagem: int = 0.95, sort : bool = True):
    """
    Compara uma string (org) com uma lista de strings (dests), retornando a mais semelhante
    de acordo com a porcentagem informada (porcentagem)
    :param org: string ser comparada
    :param dests: lista de strings a ser comparada
    :param porcentagem: porcentagem minima de igualdade entre as strings a ser considerada
    :param sort: true se as palavras devem ser ordenadas
    :return:
    """
    nome_retorno = None
    old_calc = 0
    for idx, aut in enumerate(dests):
        calc = comparador(normalizar(org, sort), normalizar(aut, sort), True)
        if calc > porcentagem and calc > old_calc:
            old_calc = calc
            nome_retorno = aut

    return nome_retorno


from __future__ import annotations

from secrets import token_hex
from typing import Any, Iterator

import psycopg2
from pandas import DataFrame
from pgcopy import CopyManager
from sqlalchemy import create_engine


class PostgresDatabase:
    def __init__(self, conn: str, table: str) -> None:
        self.__conn = conn
        self.__table = table
        self.__table_rename = f"rename_{table}_{token_hex(nbytes=8)}"
        self.__tmp_table = f"tmp_{table}_{token_hex(nbytes=8)}"
        self.__destination = psycopg2.connect(conn)
        self.__destination.autocommit = False
        self.__cursor = self.__destination.cursor(cursor_factory=psycopg2.extras.DictCursor)
        self.__cursor.itersize = 10000

    def save(self, columns, data, first=True) -> bool:
        if first:
            return self.__initial_saver(columns, data)
        return self.__copy_manager_saver(columns, data)

    def close(self) -> None:
        self.__cursor.execute(
            f"ALTER TABLE IF EXISTS {self.__table} RENAME TO {self.__table_rename}"
        )
        self.__cursor.execute(f"ALTER TABLE IF EXISTS {self.__tmp_table} RENAME TO {self.__table}")
        self.__cursor.execute(f"DROP TABLE IF EXISTS {self.__table_rename}")
        self.__cursor.execute(f"DROP TABLE IF EXISTS {self.__tmp_table}")
        self.__destination.commit()
        self.__cursor.close()
        self.__destination.close()

    def __enter__(self) -> PostgresDatabase:
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def __initial_saver(self, columns: Iterator[Any], data: Iterator[Any]) -> bool:
        try:
            # Create the table with the first itens ...
            engine = create_engine(self.__conn, echo=False)
            df = DataFrame(data, columns=columns)
            df.to_sql(f"{self.__tmp_table}", index=False, con=engine)
            engine.dispose()
        except Exception as e:
            print(e)
            return False
        return True

    def __copy_manager_saver(self, columns: Iterator[Any], data: Iterator[Any]) -> bool:
        try:
            mgr = CopyManager(self.__destination, f"{self.__tmp_table}", columns)
            mgr.copy(data)
            self.__destination.commit()
        except Exception as e:
            print(e)
            return False
        return True


# Always prefer setuptools over distutils
from setuptools import setup


# This call to setup() does all the work
setup(
    name='edat_ingestao_utils',
    version='0.0.4',
    description='Biblioteca de Apoio a Ingestão de Dados no EDAT',
    long_description='# Utilitarios EDAT <br /> Classes utilitarias de ingestão utilizadas pelo EDAT.',
    long_description_content_type='text/markdown',
    author='Escritório de Dados',
    author_email='dados@unicamp.br',
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Operating System :: OS Independent'
    ],
    packages=['edat_ingestao_utils'],
    include_package_data=True,
    install_requires=['psycopg2==2.8.6', 'SQLAlchemy==1.3.15', 'pandas==0.24.2', 'pgcopy==1.4.1', 'jellyfish==0.8.2'],
    python_requires='>=3.6'
)
